# Copyright (c) 2015 Ultimaker B.V.
# Uranium is released under the terms of the AGPLv3 or higher.

from PyQt5.QtCore import QObject, pyqtSlot, pyqtSignal,QVariant,pyqtProperty
from PyQt5.QtWidgets import QAction
from PyQt5.QtQml import QQmlListProperty,QJSValue
from UM.Scene.Iterator.DepthFirstIterator import DepthFirstIterator
from UM.Scene.Iterator.BreadthFirstIterator import BreadthFirstIterator

from UM.Scene.SceneNode import  SceneNode
from PyQt5.QtCore import Q_ENUMS
from UM.Application import Application
from UM.Scene.Selection import Selection
from UM.Scene.Camera import Camera

from UM.Operations.OperationStack import  OperationStack
from UM.Operations.AddSceneNodeOperation import AddSceneNodeOperation

from cura.ReversedPlateManager import ReversedPlateManager

from threading import Condition

import time
import copy

from cura.SIProxies.BaseNotifications import BaseNotifications
from enum import Enum



class NotificationsProxy(BaseNotifications):
    """
    Class visible in QML file as Gence.Notifications
    """

    instance = None
    condition=Condition()

    class AppState:
        EMPTY_PLATE = 0
        NODE_LOADING = 1
        NODE_LOADED = 3
        GCODE_LOADING = 4
        GCODE_LOADED = 5
        SLICING = 6
        SLICED = 7
        SAVED = 8
        ERROR_DISCONNECTED = 9
        ERROR_UNABLETOSAVE = 10
        ERROR_UNABLETOREAD = 11

    Q_ENUMS(AppState)

    stateChanged = pyqtSignal(AppState, arguments=['appState'])
    xChanged = pyqtSignal(float,arguments=['xPos'])
    yChanged = pyqtSignal(float,arguments =['yPos'])
    zChanged = pyqtSignal(float,arguments = ['zPos'])

    numberOfNodesChanged = pyqtSignal(int)
    # viewmodeChanged = pyqtSignal(bool,arguments=['perspective'])

    ######################### INIT ###########################################################################
    ##########################################################################################################
    def __init__(self, parent = None):
        super().__init__(parent)

        self._application = Application.getInstance()
        self._controller = self._application.getController()
        self._scene =self._controller.getScene()
        self._scene.sceneChanged.connect(self.notifyAllListeners)
        self._operationStack = self._application.getOperationStack()
        self._operationStack.changed.connect(self.onOperationStackChanged)
        self._backend = self._application.getBackend()
        self._reversedPlateManager = ReversedPlateManager().getInstance()

        self._xPos=0
        self._numberOfNodesBeforeGCode = 0

        # Hold number of selectable objects on scene (Models)
        self._selectableCounter=0
        # Holds name of first node on scene which is helpful while creating new name for gcode file
        self._nameOfFirstNode=None
        # Tells us if we trigger new node adding, usefull when we want to know if there were any new nodes added to scene
        # it helps to exlude some mistakes
        self._operationPerformed=False

        self._backend.processingDone.connect(self._onProcessingDone)
        self._appState= NotificationsProxy.AppState.ERROR_DISCONNECTED
        self.stateChanged.connect(self._onAppStateChanged)

        self._listOfNodes=[]

        # Filename of outuput gcode file, it is set in StatusAfterContent.qml and colud be "Group..." or
        # name from stl file
        self._outputFilename=""


    # def __del__(self):
    #     self._backend.processingDone.disconnect(self._onProcessingDone)
    #     self._controller.getScene().sceneChanged.disconnect(self.notifyAllListeners)
    #     self.stateChanged.disconnect(self._onAppStateChanged)

    ############################# MODEL MOVED ################################################################



    def _onProcessingDone(self):
        self.appState = NotificationsProxy.AppState.SLICED

    @pyqtProperty(AppState,notify=stateChanged)
    def appState(self):
        """ Keeps current application state ,
            Thanks to that we know if something in on plate,
            or something is sliced
        """
        return self._appState

    @appState.setter
    def appState(self,newState):
        if(newState != self._appState):
            self._appState = newState
            self.stateChanged.emit(self._appState)

    def setNumberOfNodesBeforeGCode(self, nodes_number):
        self._numberOfNodesBeforeGCode = nodes_number

    # Holds position of node after operation
    @pyqtProperty(float,notify=xChanged)
    def xPos(self):
        return self._xPos
    
    @xPos.setter
    def xPos(self,newX):
        self._xPos=newX
        self.xChanged.emit(newX)

    # Holds position of node after operation
    @pyqtProperty(float, notify=yChanged)
    def yPos(self):
        return self._yPos

    @yPos.setter
    def yPos(self, newy):
        self._yPos = newy
        self.yChanged.emit(newy)

    # Holds position of node after operation
    @pyqtProperty(float, notify=zChanged)
    def zPos(self):
        return self._zPos

    @zPos.setter
    def zPos(self, newz):
        self._zPos = newz
        self.zChanged.emit(newz)

    def _onAppStateChanged(self,newState):
        """
        Inform qml view about changing in application state
        should be remove because it is better to monitor property
        :param newState:
        :return:
        """
        self.getCallback("onAppStateChanged")(newState)

    @pyqtProperty(int,notify=numberOfNodesChanged)
    def numberOfNodes(self):
        return self._selectableCounter

    @pyqtProperty(str)
    def nameOfFirstNode(self):
        """
        Useful while creating gcode filename
        :return:
        """
        return self._nameOfFirstNode

    @pyqtProperty(bool)
    def isPerspective(self):
        camera = self._scene.getActiveCamera()
        if(camera):
            return camera.isPerspective()
        return False
    @isPerspective.setter
    def isPerspective(self,newState):
        camera = self._scene.getActiveCamera()
        if(camera):
            camera.setPerspective(newState)
            main_window = Application.getInstance().getMainWindow()
            main_window.updateViewportGeometry()

    outputFilenameChanged = pyqtSignal(str);

    @pyqtProperty(str,notify=outputFilenameChanged)
    def outputFilename(self):
        return self._outputFilename;

    @outputFilename.setter
    def outputFilename(self,newName):
        self._outputFilename=newName;
        self.outputFilenameChanged.emit(newName)

    def showGcode(self):
        Selection.clear()
        self._reversedPlateManager.reversePlateIfNecessary()
        self._controller.setSelectionTool("EmptyTool")
        self._controller.setActiveView("LayerView")
        view = self._controller.getActiveView()
        view.resetLayerData()
        view.setLayer(9999999)
        view.calculateMaxLayers()

    @pyqtSlot()
    def showSliced(self):
        Selection.clear()
        self.appState = NotificationsProxy.AppState.SLICING
        self._controller.setSelectionTool("EmptyTool")
        self._controller.setActiveView("LayerView")
        self._reversedPlateManager.reversePlateIfNecessary()
        self._application.getBackend().slice()

    @pyqtSlot()
    def showSolid(self, try_to_reverse = True):
        if self.appState == NotificationsProxy.AppState.GCODE_LOADED:
            if 0 < self._numberOfNodesBeforeGCode:
                Application.getInstance().getOperationStack().undo()
                self.appState = NotificationsProxy.AppState.NODE_LOADED
                self._numberOfNodesBeforeGCode = 0
            else:
                self.appState = NotificationsProxy.AppState.EMPTY_PLATE
        else:
            self.appState = NotificationsProxy.AppState.NODE_LOADED

        if try_to_reverse:
            self._reversedPlateManager.reversePlateIfNecessary()

        self._controller.setSelectionTool("SelectionTool")
        self._controller.setActiveView("SolidView")

    @pyqtSlot()
    def reversePlateIfNecessary(self):
        self._reversedPlateManager.reversePlateIfNecessary()

    @pyqtSlot()
    def arrange(self):
        print("Arranging")

    def _checkNodesAmount(self, lastAmountOfNodes: int):
        selectable_nodes = Selection.getAllSelectableNodes()
        curentAmountOfNodes = len(selectable_nodes)

        if self.appState != NotificationsProxy.AppState.GCODE_LOADED and self.appState != NotificationsProxy.AppState.GCODE_LOADING:
            if curentAmountOfNodes == 0:
                self.appState = NotificationsProxy.AppState.EMPTY_PLATE
            else:
                self._nameOfFirstNode = selectable_nodes[0]._name
                if lastAmountOfNodes < curentAmountOfNodes:
                    self.appState = NotificationsProxy.AppState.NODE_LOADED

                    self._operationPerformed=False

        self._selectableCounter=curentAmountOfNodes

    def onOperationStackChanged(self):
        self._operationPerformed=True

    def notifyAllListeners(self,node=None):
        if node:
            # node and node == self._controller.getScene().getRoot()):
            if self._operationPerformed:
                self._checkNodesAmount(self._selectableCounter)
                self._operationPerformed = False

            self.getCallback("onNodeTransformed")()

    deactivatePlaceOnBed = pyqtSignal(bool, arguments = ["activateMoveTool"])
    def deactivatePlaceOnBedTool(self, activateMoveTool = True):
        self.deactivatePlaceOnBed.emit(activateMoveTool)

    @classmethod
    def createNotificationsProxy(cls, engine = None,  script_engine = None):
        if NotificationsProxy.instance is None:
                NotificationsProxy.instance = cls()

        Application.getInstance()._notifications = NotificationsProxy.instance

        return NotificationsProxy.instance

    splittingState = pyqtSignal(str, arguments = ["state"])
    def splittingStateTool(self, state:str):
        self.splittingState.emit(state)
