import sqlite3
import os
import collections
import string
import shutil
import base64
from UM.Resources import Resources
from UM.Logger import Logger
from UM.Application import Application

class PrinterSettingsDB:
    db_file_name = "settings.db"

    def __init__(self):
        new_db = False
        db_path = os.path.join(Resources.getPath(Resources.GenceSettings), self.db_file_name)


        if not os.path.isfile(db_path):
            db_default_path = os.path.join(Resources.getPath(Resources.DefaultDatabase), self.db_file_name)
            shutil.copyfile(db_default_path, db_path)
            new_db = True

        self.db_connection = sqlite3.connect(db_path)
        self.db_cursor = self.db_connection.cursor()

        # Create copies of profiles for new database
        if new_db:
            profiles_copy = self.db_cursor.execute('SELECT id, version, printer, nozzle, material, quality, durability, profile FROM profiles_files WHERE active=1;').fetchall()

            for p in profiles_copy:
                self.create_profile_file({"id": p[0], "version": p[1], 'printer': p[2], 'nozzle': p[3],
                                          'material': p[4], 'quality': p[5], 'durability': p[6], 'profile': p[7]})

        # We want foreign_keys for printer<->profile connection
        self.db_cursor.execute('PRAGMA foreign_keys = ON;')

    def close(self):
        self.db_connection.close()

    #######################################################
    # Getting files versions for updater
    #######################################################
    def get_fdm_version(self):
        result = None
        try:
            result = self.db_cursor.execute('SELECT MAX(version) FROM settings_files').fetchone()
            result = result[0]
        except sqlite3.Error as error:
            Logger.log("e", "Error while getting FDM version from database: {0}", error)

        return result

    def get_printers_versions(self):
        return self.__get_versions("printers")

    def get_extruders_versions(self):
        return self.__get_versions("extruders")

    def get_profiles_versions(self):
        return self.__get_versions("profiles")

    def __get_versions(self, name: str):
        result = []
        try:
            result = self.db_cursor.execute('SELECT id, version FROM ' + name.lower() + '_files WHERE active=1').fetchall()
            result = [{"id": element_id, "version": version} for (element_id, version) in result]
        except sqlite3.Error as error:
            Logger.log("e", "Error while getting {0} versions from database: {1}", name.upper(), error)

        return result

    #######################################################
    # Update/Add new data to database
    #######################################################
    def update_fdm(self, update_dfm=None):
        if update_dfm is not None:
            try:
                self.db_cursor.execute('DELETE FROM settings_files')
                self.db_cursor.execute('INSERT INTO settings_files VALUES (?,?,?)', (str(update_dfm['version']), update_dfm['fdm_printer'], update_dfm['fdm_extruder']))
                self.db_connection.commit()
            except sqlite3.Error as error:
                Logger.log("e", "Error while updating FDM files in database: {0}", error)
                self.db_connection.rollback()
                return False
        return True

    def update_printers(self, active_printers, printers=None):
        result = self.__deactivate_elements("printers", active_printers)
        result = True
        if active_printers is not None:
            if printers is not None:
                for printer in printers:
                    printer['active'] = 1
                    result = self.__update_element("printers", printer)
                    if result is False:
                        break

        return result

    def update_extruders(self, active_extruders, extruders=None):
        result = True
        if active_extruders is not None:
            result = self.__deactivate_elements("extruders", active_extruders)
            if extruders is not None and result:
                for extruder in extruders:
                    extruder['active'] = 1
                    result = self.__update_element("extruders", extruder)
                    if result is False:
                        break

        return result

    def update_profiles(self, active_profiles, profiles=None):
        result = True
        if active_profiles is not None:
            result = self.__deactivate_elements("profiles", active_profiles)
            if profiles is not None and result:
                for profile in profiles:
                    profile['active'] = 1
                    profile['profile'] = base64.urlsafe_b64decode(bytes(profile['profile'], 'utf-8'))
                    result = self.__update_element("profiles", profile)
                    if result is False:
                        break
                    self.create_profile_file(profile)

        return result

    def __deactivate_elements(self, table_name, active_list):
        if active_list is not None:
            if len(active_list) != 0:
                placeholders = ','.join((['?'] * len(active_list)))

                try:
                    self.db_cursor.execute('UPDATE ' + table_name.lower() + '_files SET active=0 WHERE id NOT IN(' + placeholders + ');', active_list)
                    self.db_connection.commit()
                except sqlite3.Error as error:
                    Logger.log("e", "Error while deactivating {0} in database: {1}", table_name.upper(), error)
                    self.db_connection.rollback()
                    return False
                return True

        return False

    def __update_element(self, table, data):
        try:
            exists = self.db_cursor.execute('SELECT EXISTS(SELECT 1 FROM ' + table.lower() + '_files WHERE id=' + str(data['id']) + ' LIMIT 1);').fetchone()[0]

            keys, values = zip(*data.items())

            if exists:
                eid = data['id']
                del data['id']
                prepared_set = ['{}=?'.format(str(key)) for key in keys]
                self.db_cursor.execute('UPDATE {0}_files SET {1} WHERE id={2}'.format(table.lower(), ','.join(prepared_set), str(eid)), values)
            else:
                self.db_cursor.execute('INSERT INTO {0}_files({1}) VALUES ({2})'.format(table.lower(), ','.join(keys), ', '.join('?' for _ in values)), values)

            self.db_connection.commit()
        except sqlite3.Error as error:
            Logger.log("e", "Error while adding element into {0} table in database: {1}", table.upper(), error)
            self.db_connection.rollback()
            return False

        return True

    #######################################################
    # Getting data for file creation
    #######################################################
    def get_base_files(self):
        files = []

        fdm_files = self.__get_fdm_files()
        printers_files = self.__get_files("printers")
        extruders_files = self.__get_files("extruders")

        if fdm_files and printers_files:
            files.extend(fdm_files)
            files.extend(printers_files)
            files.extend(extruders_files)
        else:
            raise ValueError("No Printers, Extruders or FDM files")

        return files

    def __get_fdm_files(self):
        files = []
        try:
            fdm_files = self.db_cursor.execute('SELECT fdm_printer_json, fdm_extruder_json FROM settings_files ORDER BY version DESC LIMIT 1').fetchone()
            files.append({"name": "fdmprinter", "content": fdm_files[0]})
            files.append({"name": "fdmextruder", "content": fdm_files[1]})
        except sqlite3.Error as error:
            Logger.log("e", "Error while fetching FDM data from database: {0}", error)

        return files

    def __get_files(self, name: str):
        files = None
        try:
            result = self.db_cursor.execute('SELECT file_name, config FROM ' + name.lower() + '_files WHERE active=1').fetchall()
            files = [{"name": file_name.replace(" ", ""), "content": content} for (file_name, content) in result]
        except sqlite3.Error as error:
            Logger.log("e", "Error while fetching {0} data from database: {1}", name.upper(), error)

        return files



















    def get_profile(self, printer, nozzle, material, quality, durability):
        profile = self.db_cursor.execute('SELECT file_name, profile FROM profiles_files JOIN printers_files ON profiles_files.printer_id = printers_files.id WHERE printers_files.active=1 AND profiles_files.active=1 AND profiles_files.printer=? AND nozzle=? AND material=? AND quality=? AND durability=?', (printer, nozzle, material, quality, durability)).fetchall()

        return profile

    def get_elements(self, arguments: collections.OrderedDict = {}):
        first = True
        select = None
        where = None
        result = None
        where_val_list = list()

        for key, value in arguments.items():
            if value is None:
                select = key
                break
            else:
                if first:
                    where = key + "=?"
                    first = False
                else:
                    where += " AND " + key + "=?"

                where_val_list.append(value)

        if select is not None:
            if len(where_val_list) == 0:
                result = self.db_cursor.execute('SELECT DISTINCT(' + select + ') FROM profiles_files WHERE active=1 ORDER BY id ASC;').fetchall()
                # result = self.db_cursor.execute('SELECT ' + select + ', MAX(priority) FROM profiles_files WHERE active=1 GROUP BY ' + select + ' ORDER BY priority DESC;').fetchall()
            else:
                result = self.db_cursor.execute('SELECT DISTINCT(' + select + ') FROM profiles_files WHERE active=1 AND ' + where + ' ORDER BY id ASC;', where_val_list).fetchall()
                # result = self.db_cursor.execute('SELECT ' + select + ', MAX(priority) FROM profiles_files WHERE active=1 AND ' + where + ' GROUP BY ' + select + ' ORDER BY priority DESC;', where_val_list).fetchall()
        else:
            result = "empty"

        result_tuple = select, result

        return result_tuple

    def get_profile_by_id(self, profile_id: str):
        try:
            profile = self.db_cursor.execute('SELECT printer, nozzle, material, quality, durability FROM profiles_files WHERE active=1 AND id=' + str(profile_id) + ';').fetchone()
        except:
            profile = None
        return profile

    def get_default_profile_by_printer(self, printer: str):
        try:
            printer_id = self.db_cursor.execute('SELECT printer_id FROM profiles_files WHERE active=1 AND printer=?;', (printer,) ).fetchone()

            default_profile = self.db_cursor.execute('SELECT default_profile FROM printers_files WHERE active=1 AND id=?;', (str(printer_id[0]),)).fetchone()
            profile = self.get_profile_by_id(str(default_profile[0]))

        except:
            profile = None

        return profile

    def get_first_active_profile(self):
        try:
            printer_id = self.db_cursor.execute('SELECT printer_id FROM profiles_files WHERE active=1').fetchone()
            default_profile = self.db_cursor.execute('SELECT default_profile FROM printers_files WHERE active=1 AND id=?;', (str(printer_id[0]),)).fetchone()
            profile = self.get_profile_by_id(str(default_profile[0]))
        except:
            profile = None

        return profile

    def get_profile_id_from_options(self, printer: str, nozzle: str, material: str, quality: str, durability: str):
        profile_id = self.db_cursor.execute('SELECT id FROM profiles_files WHERE active=1 AND printer=? AND nozzle=? AND material=? AND quality=? AND durability=?;', (printer, nozzle, material, quality, durability)).fetchone()
        return profile_id[0]



    def format_filename(self, s):
        valid_chars = "+-_.() %s%s" % (string.ascii_letters, string.digits)
        filename = ''.join(c for c in s if c in valid_chars)
        #filename = filename.replace(' ', '_')  # I don't like spaces in filenames.

        return filename

    def create_profile_file(self, profile_data):
        path = Resources.getCustomProfilesPath()

        try:
            file_name = str(profile_data['printer']) + "_" + str(profile_data['nozzle']) + "_" + \
                        str(profile_data['material']) + "_" + str(profile_data['quality']) + "_" +\
                        str(profile_data['durability']) + "_ID" + str(profile_data['id']) + "ver" + \
                        str(profile_data['version'])
            file_name = self.format_filename(file_name)
            file_name = path + os.sep + file_name

            # Add number to the name if file with same name already exists
            if os.path.isfile(file_name + ".profile"):
                i = 0

                while True:
                    temp_file_name = file_name + "-" + str(i) + ".profile"
                    i = i + 1
                    if not os.path.isfile(temp_file_name):
                        file_name = temp_file_name
                        break
            else:
                file_name = file_name + ".profile"

            file_obj = open(file_name, "wb")
            file_obj.write(profile_data['profile'])
            file_obj.close()
        except IOError as e:
            Logger.log("e", "Error while creating profiles (during user database creation): %s %s", e.errno, e.strerror)
        except:
            Logger.log("e", "Error while creating profiles (during user database creation)")

    def get_printer_id_by_real_name(self, printer_file_name: str):
        try:
            default_id = self.db_cursor.execute('SELECT id FROM printers_files  WHERE active=1 AND file_name=?', (printer_file_name, )).fetchone()

            # No such printer
            if default_id is None:
                return False

            return default_id[0]

        except:
            return False

    def get_default_profile_by_printer_id(self, printer_id: str):
        try:
            default_profile_id = self.db_cursor.execute('SELECT default_profile FROM printers_files  WHERE active=1 AND id=?', (str(printer_id), )).fetchone()

            if default_profile_id is None:
                return False

            profile = self.get_profile_by_id(default_profile_id[0])

            if profile is None:
                return False

            return profile
        except:
            return False

    def get_any_profile_by_printer_id(self, printer_id: str):
        try:
            profile_id = self.db_cursor.execute('SELECT id FROM profiles_files  WHERE active=1 AND printer_id=?', (str(printer_id), )).fetchone()

            if profile_id is None:
                return False

            profile = self.get_profile_by_id(profile_id[0])

            if profile is None:
                return False

            return profile
        except:
            return False


    def get_friendly_printer_name(self, name: str):
        friendly_name = None

        try:
            default_id = self.db_cursor.execute('SELECT default_profile, id FROM printers_files  WHERE active=1 AND file_name=?', (name, )).fetchone()

            # No such printer
            if default_id is None:
                return False

            # Check default profile for this printer
            if default_id[0] is not None:
                friendly_name = self.db_cursor.execute('SELECT printer FROM profiles_files  WHERE id=?', (default_id[0],)).fetchone()

            # Fallback when default_profile points to nothing
            if friendly_name is None:
                friendly_name = self.db_cursor.execute('SELECT printer FROM profiles_files  WHERE printer_id=?', (default_id[1],)).fetchone()

            # There are no profiles with this printer
            if friendly_name is None:
                return False
        except:
            return False

        return str(friendly_name[0])

    def is_printer_mirrored(self):
        current_printer_name = Application.getInstance().getMachineManager().activeMachineName

        try:
            mirrored_bed = self.db_cursor.execute('SELECT mirrored_bed FROM printers_files  WHERE active=1 AND file_name=?', (current_printer_name, )).fetchone()
            mirrored_bed = mirrored_bed[0]
        except:
            return None

        if mirrored_bed is None:
            return None
        if mirrored_bed == 1:
            return True
        if mirrored_bed == 0:
            return False

        return None
