from cura import CuraVersion
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot
from UM.Preferences import Preferences
from cura.Updater.ServerClient import ServerClient
from cura.SIProxies.SettingsManage.PrinterSettingsDB import PrinterSettingsDB
from .Status import Status
from .UpdateError import UpdateError
from .UpdateData import UpdateData

class Updater(QObject):

    updateError = UpdateError.NoneError
    __client = None
    __pathToDataToUpdate = None

    # Class constructor
    def __init__(self, root):
        super().__init__()
        Preferences.getInstance().addPreference("cura/isUpdateWasInstalled", False)
        self.__root = root


    # Set update error
    @staticmethod
    def set_update_error(_updateError):
        Updater.updateError = _updateError

    # Get update error
    @staticmethod
    def get_update_error():
        return Updater.updateError

    #######################################################################################

    # Signals
    __startUpdateSignal = pyqtSignal()
    __checkIfNewAppVersionIsAvailableSignal = pyqtSignal()
    __downloadUpdate = pyqtSignal()

    # Check for update on application start
    def check_for_update_on_app_start(self):
        is_update_active = self.__check_if_auto_update_active()
        is_update_was_installed = self.__check_if_update_was_installed()
        status = Status()
        error_code = self.get_update_error()

        if error_code == UpdateError.NoneError:
            if is_update_was_installed is True:
                self.__printMessage (status.get_status_description(12))
            else:
                if is_update_active is True:
                    self.__start_update()
                else:
                    self.__printMessage (status.get_status_description(4))
        else:
            self.__printMessage(status.get_status_description(16,error_code.value))

    # Display message 'Checking for updates...'
    def __start_update (self):
        status = Status()
        self.__printMessage(status.get_status_description(1))
        self.__download_update_slot()

    # Download update
    def __download_update_slot (self):
        applicationVersion = CuraVersion.CuraVersion
        printerSettingsDB = PrinterSettingsDB()
        activeProfiles = printerSettingsDB.get_profiles_versions()
        activePrinters = printerSettingsDB.get_printers_versions()
        activeExtruders = printerSettingsDB.get_extruders_versions()
        fdmVersion = printerSettingsDB.get_fdm_version()
        client = ServerClient()

        client.setArgs([applicationVersion, activeProfiles, activePrinters, activeExtruders, fdmVersion])
        client.start()

    # Display message.
    @pyqtSlot(int)
    def display_message_slot(self, status):
        _status = Status()
        self.__printMessage(_status.get_status_description (status))

    # Install update.
    def install_update(self):
        update_data = UpdateData()
        if update_data.file_exists() is True:
            update_data.load_from_file()
            if self.__update_database(update_data):
                update_data.remove_file()
                Preferences.getInstance().setValue("cura/isUpdateWasInstalled", True)
            else:
                Preferences.getInstance().setValue("cura/isUpdateWasInstalled", False)

    # Check for update on button click.
    def check_for_update_on_click(self):
        if not UpdateData().file_exists():
            self.__start_update()

    # Check whether auto update active.
    def __check_if_auto_update_active(self):
        auto_update_active = Preferences.getInstance().getValue("cura/is_update_active")
        return True if auto_update_active is True else False

    # Check whether update was installed.
    def __check_if_update_was_installed(self):
        is_update_was_installed = Preferences.getInstance().getValue("cura/isUpdateWasInstalled")
        return True if is_update_was_installed is True else False

    # Display message on state panel.
    def __printMessage(self, msg):
        self.__root.displayStatus (msg)

    # Update database.
    def __update_database(self, data):

        printer_settings_db = PrinterSettingsDB()

        result = printer_settings_db.update_printers(data.get_active_printers(), data.get_new_printers())

        if result is True:
            result = printer_settings_db.update_extruders(data.get_active_extruders(), data.get_new_extruders())
        else:
            Updater.set_update_error(UpdateError.AddPrintersError)
            return False

        if result is True:
            result = printer_settings_db.update_profiles(data.get_active_profiles(), data.get_new_profiles())
        else:
            Updater.set_update_error(UpdateError.AddExtrudersError)
            return False

        if result is True:
            result = printer_settings_db.update_fdm(data.get_new_fdm())
        else:
            Updater.set_update_error(UpdateError.AddProfilesError)
            return False

        if result is True:
            return result
        else:
            Updater.set_update_error(UpdateError.AddFdmError)