from UM.Operations.AddSceneNodeOperation import AddSceneNodeOperation
from UM.Operations.RemoveSceneNodeOperation import RemoveSceneNodeOperation
from cura.SIProxies.NotificationsProxy import NotificationsProxy
from UM.Operations.GroupedOperation import GroupedOperation
from UM.Mesh.MeshBuilder import MeshBuilder
from UM.Scene.SceneNode import SceneNode
from UM.Scene.Selection import Selection
from UM.Application import Application
from UM.Math.Vector import Vector
from UM.Job import Job
import numpy

class Side:
    Left = 0
    Right = 1
    OnPlane = 2
    Cut = 100

class Axis:
    X = 0
    Y = 1
    Z = 2

class Split(Job):
    def __init__(self):
        super().__init__()
        self.__progress = 0
        self.__nodes_count = 0

        self._controller = Application.getInstance().getController()
        self._scene = self._controller.getScene()
        self._notifications = NotificationsProxy.createNotificationsProxy()

    # Initialize progress so you don't have to worry about nodes count while calculating progress
    def _initialize_progress(self, num_of_nodes, message=None):
        self._notifications.splittingStateTool("Splitting")
        self.__progress = 0
        self.__nodes_count = num_of_nodes
        if message:
            self._display_status(message)
        Application.getInstance().getBackend().processingProgress.emit(0)

    # Adds to progress part of whole operation addends sum should be equal 1
    def _add_progress(self, addend):
        self.__progress += addend
        if self.__progress > 1:
            self.__progress = 1
        Application.getInstance().getBackend().processingProgress.emit(self.__progress)

    # Adds to progress part of whole operation divided by nodes_count - addends sum should be equal 1
    def _add_node_progress(self, addend):
        self.__progress += addend / self.__nodes_count
        if self.__progress > 1:
            self.__progress = 1
        Application.getInstance().getBackend().processingProgress.emit(self.__progress)

    # Make sure that progressbar always fill up
    def _finish_progress(self, succ = True):
        if succ:
            self._notifications.splittingStateTool("Success")
        else:
            self._notifications.splittingStateTool("Fail")
        Application.getInstance().getBackend().processingProgress.emit(1)

    # s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,s2n+1,s2n+2,...s3n-1), ...
    def _grouped(self, iterable, n):
        return zip(*[iter(iterable)] * n)

    # Comprasion for ndarrays, we can use fuzzy logic by swaping only this function
    def _v_eq(self, one, two):
        return (one == two).all()

    def __add_node(self, verticies, name="Unknown"):
        if not verticies.any():
            return

        mesh_builder = MeshBuilder()
        mesh_builder.setFileName("Color it for me, please.")
        mesh_builder.setVertices(verticies)

        mesh_builder.calculateNormals(fast=True)
        mesh = mesh_builder.build()

        scene_node = SceneNode()

        scene_node.setMeshData(mesh)
        scene_node.setSelectable(True)
        scene_node.setName(name)

        center = Vector()
        for i in range(3):
            minimumum = numpy.amin(mesh._vertices[:, i])
            maximumum = numpy.amax(mesh._vertices[:, i])
            center._data[i] = maximumum - abs(maximumum - minimumum) / 2

        scene_node.setCenterPosition(center)
        scene_node.translate(center, SceneNode.TransformSpace.World)

        return AddSceneNodeOperation(scene_node, self._scene.getRoot())

    def _get_new_node_name(self, name, number):
        splitted_name = name.split(".")
        if len(splitted_name) >= 3:
            if splitted_name[-3] != "part":
                splitted_name.insert(len(splitted_name) - 1, "part." + str(number))
        else:
            splitted_name.insert(len(splitted_name) - 1, "part." + str(number))
        new_name = '.'.join(map(str, splitted_name))
        return new_name

    def _swap_nodes_operation(self, node, left_part, right_part):
        # Prepare operations
        left_op = self.__add_node(numpy.asarray(left_part, dtype=numpy.float32), self._get_new_node_name(node.getName(), 1))
        right_op = self.__add_node(numpy.asarray(right_part, dtype=numpy.float32), self._get_new_node_name(node.getName(), 2))

        if not left_op or not right_op:
            return  # Whole model is on one side, no need to split

        # Go!
        Selection.clear()

        grouped_operation = GroupedOperation()
        grouped_operation.addOperation(left_op)
        grouped_operation.addOperation(right_op)
        grouped_operation.addOperation(RemoveSceneNodeOperation(node))
        grouped_operation.push()

        # Select newly added nodes
        selectableNodes = Selection.getAllSelectableNodes()

        if 2 <= len(selectableNodes):
            Selection.add(selectableNodes[-1])
            Selection.add(selectableNodes[-2])

    # Display status message
    def _display_status(self, msg):
        Application.getInstance().displayStatus(msg)

