import QtQuick 2.0
import QtQuick.Layouts 1.1

import "../../singletons"
import UM 1.2 as UM
import Gence 1.0 as Gence
import Cura 1.0 as Cura

RowLayout {

    property int state: UM.Backend.state

    id: message
    property variant printDuration: Printer.currentPrintTime
    property variant amount: Printer.currentAmount
    property variant printMaterialLengths: Printer.materialLengths
    property variant printMaterialWeights: Printer.materialWeights
    property string updateStatusMsg: Printer.updateStatusMsg
    property string actioName: ""
    property bool showDetails: false
    property bool showSupport: false
    property int maximumFileNameLength

    Item {
        id: font_size
        property double basic: Globals.fontPixelSize
        property double small: Globals.statusFontPixelSize
        property double big: Globals.statusBigFontPixelSize
        property double header: Globals.statusHeaderFontPixelSize
    }


    //------------------------------------------------------------------------------------------------ CONNECTIONS

    //Called on ----------------------------------------------- [ Selection Changed ]
    Connections {
        id: selectionConnection
        target: UM.Selection
        enabled: Gence.Notifications.appState === Gence.Notifications.NODE_LOADED
        onHasSelectionChanged: {
            if (UM.Selection.hasSelection) {
                showDetails = true
                setModelName(UM.Selection.selectedNode, "")
                var nodeSize = UM.Selection.selectedSize
                var x = nodeSize[0]
                var y = nodeSize[1]
                var z = nodeSize[2]
                setDimensionLabel(x, y, z)
                setVolumeLabel(x, y, z)
            } else {
                showDetails = false
            }
        }
    }


    //When slicing canceled clear status text.
    Connections {
        target: Gence.Actions.abnormalUndo
        onTriggered: {
            showDetails = false
            status.text = ''
            volume.text = ''
            size.text = ''
            name.text = ''
            form.fill.width = 0
        }
    }


    //Called on ----------------------------------------------- [ App state changed ]
    Connections {
        target: Gence.Notifications
        onAppStateChanged: {
            var appState = Gence.Notifications.appState

            switch (appState) {
            case Gence.Notifications.SLICED:
                showDetails = true
                var newName = ""
                if (Gence.Notifications.numberOfNodes > 1) {
                    newName = new Date().toLocaleString(Qt.locale(), "Group_yyyy-MM-dd_hh-mm")
                } else {
                    var singleNodeName = Gence.Notifications.nameOfFirstNode
                    singleNodeName = singleNodeName.substr(0, singleNodeName.lastIndexOf("."))
                    newName = singleNodeName
                }
                var nodesBox = UM.Selection.nodesSize
                setModelName(newName, ".gcode")
                break
            case Gence.Notifications.NODE_LOADED:
                showDetails = true
                status.text = ''
                actionLabel.text = "Node loaded"
                form.fill.width = 0
                break
            case Gence.Notifications.SLICING:
                showDetails = false
                status.text = ''
                actionLabel.text = "Slicing..."
                break
            case Gence.Notifications.NODE_LOADING:
                showDetails = false
                status.text = ''
                actionLabel.text = "Loading..."
                break
            case Gence.Notifications.GCODE_LOADING:
                showDetails = false
                status.text = ''
                actionLabel.text = "Parsing GCode..."
                break
            case Gence.Notifications.GCODE_LOADED:
                showDetails = false
                status.text = ''
                actionLabel.text = "GCode loaded"
                form.fill.width = 0
                break
            case Gence.Notifications.EMPTY_PLATE:
                showDetails = false
                status.text = ''
                actionLabel.text = ''
                form.fill.width = 0
                break
            }
        }
    }

    //Called on ------------------------------------------------- [ Material value changed ]

    //Backend(with help of GenceAplication as Proxy) has changed the amount of material in cm3
    Connections {
        target: message
        onAmountChanged: {
            volume.text = "V:<b>" + amount + "</b> cm3"

            setVolumeLabel(amount)
            if (amount !== 0) {
                actionLabel.text = ""
            }

            setMaterialUsage(printMaterialWeights[0], printMaterialLengths[0])
            setSupportUsage(printMaterialWeights[1], printMaterialLengths[1])
        }
    }

    //Called on ------------------------------------------------- [ Update status changed ]

    //Backend(with help of GenceAplication as Proxy) has changed the current update status
    Connections {
        target: CuraApplication
        onDisplayNewStatus: {
            var str = updateStatusMsg
            if (str.length !== 0) {
                status.text = str
            }
        }
    }

    //Called on ------------------------------------------------- [ Weights of used material changed - last signal from backend ]
    Connections {
        target: message
        onPrintMaterialWeightsChanged: {
            var hours = printDuration.days * 24 + printDuration.hours
            centralText.text = "<b>" + hours + "</b>h : <b>" + printDuration.minutes + "</b>m"
        }
    }

    //--------------------------------------------------------------------------------------------------------- Functions

    function setMaterialUsage(grams, meters) {
        modelUsage.text = "Model material: " + grams.toFixed(2) + "g (~" + meters.toFixed(2) + "m)"
    }

    function setSupportUsage(grams, meters) {
        if (grams === 0 || grams === undefined) {
            showSupport = false
            return
        }
        showSupport = true
        supportUsage.text = "Support material: " + grams.toFixed(2) + "g (~" + meters.toFixed(2) + "m)"
    }

    function setDimensionLabel(x, y, z) {
        size.text = x.toFixed(2) + " x " + y.toFixed(2) + " x " + z.toFixed(2) + " mm"
    }

    function setVolumeLabel(x, y, z) {
        if (!isNaN(y)) {
            volume.text = "V: " + (x / 10 * y / 10 * z / 10).toFixed(2) + " cm³"
        } else {
            volume.text = "V: " + x.toFixed(2) + " cm3"
        }
    }

    function setModelName(modelName, extension) {
        name.text = modelName + extension
        Gence.Notifications.outputFilename = modelName
    }

    //--------------------------------------------------------------------------------------------------------- GUI

    //Action text : loading, slicing
    Text {
        id: actionLabel
        visible: !showDetails
        color: Colors.statusBarText
        font.family: Fonts.boxed
        font.pixelSize: font_size.small
        anchors.verticalCenter: parent.verticalCenter
        Layout.margins: parent.width * 0.01
    }

    // //Text containing label and filename (for example:   Name : model.stl)
    // Item {
    //     id: left_container
    //     visible: showDetails
    //     anchors.verticalCenter: parent.verticalCenter
    //     anchors.left: parent.left
    //     anchors.right: centralText.left
    //     anchors.leftMargin: parent.width * 0.01
    //     anchors.rightMargin: parent.width * 0.01

    Text {
        id: labelName
        text: "Name: "
        visible: showDetails
        color: Colors.statusBarText
        font.family: Fonts.boxed
        font.pixelSize: font_size.small
        anchors.verticalCenter: parent.verticalCenter
        anchors.left: parent.left
        anchors.leftMargin: parent.width * 0.01
    }

    Text {
        id: name
        visible: showDetails
        Layout.preferredWidth: {
            var widthRatio = centralText.visible ? 0.3 : 0.4
            return hackTextItem.contentWidth > parent.width * widthRatio ? parent.width * widthRatio : hackTextItem.contentWidth
        }
        color: Colors.statusBarText
        font.family: Fonts.boxed
        font.pixelSize: font_size.small
        anchors.verticalCenter: parent.verticalCenter
        anchors.left: labelName.right
        elide: Text.ElideMiddle

        Text {
            // used to check text width to set maximum width
            id: hackTextItem
            visible: false
            text: name.text
            font {
                family: Fonts.boxed
                pixelSize: font_size.small
            }
        }
    }

    Text {
        id: size
        visible: showDetails && Gence.Notifications.appState !== Gence.Notifications.SLICED
        color: Colors.statusBarText
        font.family: Fonts.boxed
        font.pixelSize: font_size.small
        anchors.left: name.right
        anchors.verticalCenter: parent.verticalCenter
        anchors.leftMargin: parent.width * 0.01
        anchors.rightMargin: parent.width * 0.01
    }

    //Textbox containing volume of used material
    Text {
        id: volume
        visible: showDetails
        color: Colors.statusBarText
        font.family: Fonts.boxed
        font.pixelSize: font_size.small
        anchors.left: centralText.visible ? undefined : size.right
        anchors.right: centralText.visible ? centralText.left : undefined
        anchors.verticalCenter: parent.verticalCenter
        anchors.leftMargin: parent.width * 0.01
        anchors.rightMargin: parent.width * 0.01
    }
    // }

    //Central text containing calculated time of priniting
    Text {
        id: centralText
        visible: showDetails && Gence.Notifications.appState === Gence.Notifications.SLICED
        width: parent.width * 0.2
        color: Colors.statusBarText
        font.family: Fonts.boxed
        font.pixelSize: font_size.header
        anchors.horizontalCenter: parent.horizontalCenter
    }

    Item {
        id: material
        visible: showDetails && Gence.Notifications.appState === Gence.Notifications.SLICED
        width: parent.width * 0.4
        anchors.verticalCenter: parent.verticalCenter
        anchors.left: centralText.right
        anchors.leftMargin: parent.width * 0.01

        RowLayout {
            spacing: labelName.width * 0.5

            anchors.verticalCenter: parent.verticalCenter
            Text {
                id: modelUsage
                text: "Model material: 125g (~345m)"
                color: Colors.statusBarText
                font.family: Fonts.boxed
                font.pixelSize: font_size.small
            }

            Text {
                id: supportUsage
                visible: showSupport
                text: "Support material: 125g (~345m)"
                color: Colors.statusBarText
                font.family: Fonts.boxed
                font.pixelSize: font_size.small
            }
        }
    }

    //Right text containing current update status
    Text {
        id: status
        width: parent.width * 0.3
        textFormat: Text.PlainText
        color: Colors.statusBarText
        font.family: Fonts.boxed
        font.pixelSize: font_size.small
        anchors.verticalCenter: parent.verticalCenter
        anchors.right: parent.right
        anchors.margins: parent.width * 0.01
    }
}
