import QtQuick 2.4
import QtQuick.Dialogs 1.2

import "../singletons"

import Gence 1.0 as Gence
import Cura 1.0 as Cura

ModalWindowForm {
    property int animationDuration: 100
    // used to remember last used source, we don't reset position and size if someone is still using the same window.
    property string lastSetSource: ""

    signal opened()
    signal closed()
    signal closing()
    signal forcedClose()

    function close(){
        if(modalWindow.visible === true){
            closing()
            onCloseAnimation.start()
        }
    }

    function forceClose(){
        forcedClose()
        close()
    }

    Connections {
        target: onCloseAnimation
        onStopped: {
            modalWindow.visible = false
            Gence.Actions.modalWindowClosed.trigger()
            closed()
        }
    }

    function open(){
        if(modalWindow.visible === false){
            modalWindow.visible = true
            onOpenAnimation.start()
            Gence.Actions.modalWindowOpened.trigger()
            opened()
        }
    }

    function setSource(source){
        if (lastSetSource.localeCompare(source)){
        windowContent = source
            modalWindow.resizable = false
            resetAndCenter()
            lastSetSource = source
        }
    }

    function expandHeightBy(pixelsCount){
        var factorToPixels = modalWindow.height / heightFactor
        var pixelsToFactor = pixelsCount / factorToPixels
        var newHeightFactor = heightFactor + pixelsToFactor
        var newHeight = factorToPixels * newHeightFactor
        var maxY = base.height - newHeight

        if (modalWindow.y > maxY){
            modalWindow.userY = yPosToUserY(maxY)
        }

        heightFactor = newHeightFactor
    }

    function yPosToUserY(y){
        return y / base.height
    }

    function resetAndCenter(){
        // move window to left upper corner to prevent size change if bigger window
        // is loaded after small one
        modalWindow.userX = 0
        modalWindow.userY = 0

        // set default size ratio
        modalWindow.userWidth = 1
        modalWindow.userHeight = 1

        // finally center window in application window
        modalWindow.userX = 0.5 - modalWindow.width/(2*base.width)
        modalWindow.userY = 0.5 - modalWindow.height/(2*base.height)
    }

    function setTitle(title){
        windowTitle = title
    }

    function setResizable(state){
        modalWindow.resizable = state
    }

    ParallelAnimation{
        id: onOpenAnimation
        PropertyAnimation {
            target: modalWindow
            property: "opacity"
            easing.type: Easing.InCubic
            duration: animationDuration
            from: 0.0
            to: 1.0
            running: false
        }

        ColorAnimation {
            target: modalWindowBorder
            property: "color"
            easing.type: Easing.InCubic
            duration: animationDuration
            from: Colors.buttonMenuBackground
            to: Colors.buttonMenuBorder
            running: false
        }

        ScaleAnimator {
            target: modalWindow
            easing.type: Easing.InCubic
            from: 0.9
            to: 1.0
            duration: animationDuration
            running: false
        }
    }

    ParallelAnimation {
        id: onCloseAnimation
        PropertyAnimation {
            target: modalWindow
            property: "opacity"
            easing.type: Easing.OutCubic
            duration: animationDuration
            from: 1.0
            to: 0.0
            running: false
        }

        ColorAnimation {
            target: modalWindowBorder
            property: "color"
            easing.type: Easing.OutCubic
            duration: animationDuration
            from: Colors.buttonMenuBorder
            to: Colors.buttonMenuBackground
            running: false
        }

        ScaleAnimator {
            target: modalWindow
            easing.type: Easing.OutCubic
            from: 1.0
            to: 0.9
            duration: animationDuration
            running: false
        }
    }

    Connections {
        target: Printer
        onFreshInstallation: openAboutWindow()
    }

    function openAboutWindow(){
        setSource("about/About.qml")
        setTitle("About")
        open()
    }

    function customSettingsWarning() {
        if(Printer.inCustom){
            openCustomSettingsWindow()
        } else {
            customSettingsWarningDialog.open()
        }
    }

    function openCustomSettingsWindow() {
        var result = true
        if(!Printer.inCustom){
            result = Gence.CustomSettingsActions.loadProfileDialog()
        }
        if (result == true) {
            setSource("customSettings/SettingView.qml")
            setTitle("Custom Settings")
            modalWindow.resizable = true
            open()
        } else {
            Printer.displayStatus("Incorrect custom profile.")
        }
    }

    MessageDialog {
        id: customSettingsWarningDialog
        title: "Custom settings warning"
        icon: StandardIcon.Warning
        text: "Attention: these advanced options are intended to modify printing profiles for experienced users only.\n\n" +
            "Please read manual for further details. You are using these options AT YOUR OWN RISK, without any warranties. " +
            "Some modifications may damage the print, the printing module or the printer - 3DGence takes no resposibility " +
            "for damage caused by inappropriate use of these settings."
        standardButtons: StandardButton.Cancel | StandardButton.Ok
        onAccepted: openCustomSettingsWindow()
    }

    Connections {
        target: Cura.Actions.openCustomSettings
        onTriggered: customSettingsWarning()
    }
}