# Copyright (c) 2015 Ultimaker B.V.
# Uranium is released under the terms of the AGPLv3 or higher.

from . import SceneNode

from UM.Math.Matrix import Matrix
from UM.Math.Ray import Ray
from UM.Math.Vector import Vector

import numpy
import numpy.linalg
import copy


from UM.Mesh.MeshBuilder import MeshBuilder
from UM.Math.Color import Color




##  A SceneNode subclass that provides a camera object.
#
#   The camera provides a projection matrix and its transformation matrix
#   can be used as view matrix.
class Camera(SceneNode.SceneNode):
    def __init__(self, name, parent = None):
        super().__init__(parent)
        self._name = name
        self._projection_matrix = Matrix()
        self._projection_matrix.setOrtho(-5, 5, 5, -5, -100, 100)
        self._perspective = False
        self._viewport_width = 0
        self._viewport_height = 0
        self._window_width = 0
        self._window_height = 0
        self._zoom = 0.0

        self._perspectiveMatrix = None

        self.setCalculateBoundingBox(False)



    def getZoom(self):
        return self._zoom;

    def setZoom(self,newZoom):
        self._zoom=newZoom;

    def setPerspectiveMatrix(self,matrix):
        self._perspectiveMatrix=copy.deepcopy(matrix)

    def getPerspectiveMatrix(self):
        return self._perspectiveMatrix

    ##  Get the projection matrix of this camera.
    def getProjectionMatrix(self):
        return copy.deepcopy(self._projection_matrix)
    
    def getViewportWidth(self):
        return self._viewport_width
    
    def setViewportWidth(self, width):
        self._viewport_width = width
    
    def setViewPortHeight(self,height):
        self._viewport_height = height
        
    def setViewportSize(self,width,height):
        self._viewport_width = width
        self._viewport_height = height
    
    def getViewportHeight(self):
        return self._viewport_height

    def setWindowSize(self, w, h):
        self._window_width = w
        self._window_height = h
    
    ##  Set the projection matrix of this camera.
    #   \param matrix The projection matrix to use for this camera.
    def setProjectionMatrix(self, matrix):
        self._projection_matrix = matrix

    def isPerspective(self):
        return self._perspective

    def setPerspective(self, pers):
        self._perspective = pers



    ##  Get a ray from the camera into the world.
    #
    #   This will create a ray from the camera's origin, passing through (x, y)
    #   on the near plane and continuing based on the projection matrix.
    #
    #   \param x The X coordinate on the near plane this ray should pass through.
    #   \param y The Y coordinate on the near plane this ray should pass through.
    #
    #   \return A Ray object representing a ray from the camera origin through X, Y.
    #
    #   \note The near-plane coordinates should be in normalized form, that is within (-1, 1).

    def getRay(self, x, y):

        window_x = ((x + 1) / 2) * self._window_width
        window_y = ((y + 1) / 2) * self._window_height
        view_x = (window_x / self._viewport_width) * 2 - 1
        view_y = (window_y / self._viewport_height) * 2 - 1


        inverted_projection = numpy.linalg.inv(self._projection_matrix.getData().copy())
        transformation = self.getWorldTransformation().getData()


        near = numpy.array([view_x, -view_y, -1.0, 1.0], dtype = numpy.float32)
        near = numpy.dot(inverted_projection, near)
        near = numpy.dot(transformation, near)
        near = near[0:3] / near[3]

        far = numpy.array([view_x, -view_y, 1.0, 1.0], dtype = numpy.float32)
        far = numpy.dot(inverted_projection, far)
        far = numpy.dot(transformation, far)
        far = far[0:3] / far[3]

        dir = far-near
        dir /= numpy.linalg.norm(dir)


        if(not self.isPerspective()):

            pointPosition = numpy.array([view_x , -view_y , 0.0 , 1.0])
            pointPosition = numpy.dot(inverted_projection,pointPosition)
            pointPosition= numpy.dot(transformation, pointPosition)
            pointPosition = pointPosition[0:3] / pointPosition[3]
            cameraPosition=Vector(pointPosition[0],pointPosition[1],pointPosition[2])
            return Ray(cameraPosition, Vector(dir[0], dir[1], dir[2]))
        else:
            return Ray(self.getWorldPosition(), Vector(-dir[0], -dir[1], -dir[2]))

    ##  Project a 3D position onto the 2D view plane.
    def project(self, position):

        projection = self._projection_matrix
        view = self.getLocalTransformation().getInverse()
        position = position.preMultiply(view)
        position = position.preMultiply(projection)


        return position.x / position.z / 2.0, position.y / position.z / 2.0,position.z



    def calculatePointsBeforeBeingDotProjections(self,secondary):

        for point,array in enumerate(secondary):
            xScreen = secondary[point][0] * secondary[point][2] * 2
            yScreen = secondary[point][1] * secondary[point][2] * 2
            zScreen = secondary[point][2]
            position = numpy.linalg.lstsq(self._projection_matrix.getData(),
                                          numpy.array([xScreen, yScreen, zScreen, 0.0], dtype=numpy.float32))
            p0 = position[0]
            p3 = position[3]
            position = numpy.array([p0[0], p0[1], p0[2] + p0[3] + p3[3]], dtype=numpy.float32)
            secondary[point]=position



    def calculateTrans(self,primary:numpy.array,secondary:numpy.array):

        self.calculatePointsBeforeBeingDotProjections(secondary)

        import numpy as np

        n = primary.shape[0]
        pad = lambda x: np.hstack([x, np.ones((x.shape[0], 1))])
        unpad = lambda x: x[:, :-1]
        X = pad(primary)
        Y = pad(secondary)

        # Solve the least squares problem X * A = Y
        # to find our transformation matrix A
        A, res, rank, s = np.linalg.lstsq(X, Y)

        transform = lambda x: unpad(np.dot(pad(x), A))


        newTransformation = Matrix(A)
        newTransformation= newTransformation.getInverse()


        return newTransformation;


    def calculateMatchingTransformation(self,positionScreen:Vector,positionWorld:Vector)-> Matrix:
        """
         We want to get camera transformation which provide us that some point in the world coordinates,
         will be placed in some point on the screen.

        :param positionScreen: Position which we want to achive on the screen
        :param positionWorld: Position of world point which we want to be in screen (pointScreen) position;
        :return cameraLocalTransformation
        """

        xScreen = positionScreen.x * positionScreen.z *2
        yScreen = positionScreen.y * positionScreen.z *2
        zScreen = positionScreen.z

        position = numpy.linalg.lstsq(self._projection_matrix.getData(),numpy.array([xScreen,yScreen,zScreen,1.0],dtype=numpy.float32))

        p0 = position[0]
        p3 = position[3]
        position=numpy.array([p0[0],p0[1],p0[2]+p0[3]+p3[3],1.0],dtype=numpy.float32)
        position = Vector(data=position)

        oldTransformation = self.getLocalTransformation().getInverse().getData()

        xK = position.x - oldTransformation[0][0] * positionWorld.x - oldTransformation[0][1]*positionWorld.y - oldTransformation[0][2]*positionWorld.z
        yK = position.y - oldTransformation[1][0] * positionWorld.x - oldTransformation[1][1]*positionWorld.y - oldTransformation[1][2]*positionWorld.z
        zK = position.z - oldTransformation[2][0] * positionWorld.x - oldTransformation[2][1]*positionWorld.y - oldTransformation[2][2]*positionWorld.z

        oldTransformation[0][3] = xK
        oldTransformation[1][3] = yK
        oldTransformation[2][3] = zK

        return Matrix(oldTransformation).getInverse()

        return Matrix()




    def getMatrixForPosition(self,position:Vector):
        pass





