# Copyright (c) 2015 Ultimaker B.V.
# Uranium is released under the terms of the AGPLv3 or higher.

from UM.Tool import Tool
from UM.Event import Event, MouseEvent, KeyEvent
from UM.Math.Vector import Vector
from UM.Math.Quaternion import Quaternion

from UM.Math.Matrix import Matrix
from UM.Application import Application
from PyQt5 import QtCore, QtWidgets

from UM.Scene.SceneNode import  SceneNode
from UM.Mesh.MeshBuilder import MeshBuilder
from UM.Math.Color import Color

import numpy
import math
import copy

from threading import Thread, Lock


##  Provides the tool to manipulate the camera: moving, zooming and rotating
#
#   Note that zooming is performed by moving closer to or further away from the origin ("dolly")
#   instead of changing the field of view of the camera ("zoom")
class CameraTool(Tool):
    def __init__(self):
        super().__init__()
        self.application = Application.getInstance()

        self._scene = self.application.getController().getScene()
        self._mutex = Lock()
        self._yaw = 0
        self._pitch = 0
        self._origin = Vector(0, 0, 0)
        self._min_zoom = 0
        self._max_zoom = 2000.0
        self._manual_zoom = 200
        self._zoom_speed = 2
        self._current_zoom = None

        self._rotate = False
        self._move = False
        self._dragged = False

        self._shift_is_active = None
        self._ctrl_is_active = None
        self._space_is_active = None

        self._start_drag = None
        self._start_y = None

        self._drag_distance = 0.05

        self._coordinate_system = None
        self._current_size = 0
        self._zDistance=0
        self._lastTransformation=None

        self._leftPoint = numpy.array([0, 0, 0], dtype=numpy.float32)
        self._rightPoint = numpy.array([10., 0, 0], dtype=numpy.float32)
        self._max_size = 0.5
        self._min_size = 0.0039

        # root = self._scene.getRoot()
        # self._node = SceneNode(root)

        # mb = MeshBuilder()
        #
        # mb.addCube(5,5,5,Vector(0,0,0))
        # mb.addCube(5, 5, 5, Vector(10, 0, 0),Color(255,0,0,255))
        #
        #
        # self._node.setMeshData(mb.build())
        # self._node.setVisible(False)


    ##  Set the minimum and maximum distance from the origin used for "zooming" the camera
    #
    #   \param min type(float) distance from the origin when fully zoomed in
    #   \param max type(float) distance from the origin when fully zoomed out
    def setZoomRange(self, min, max):
        self._min_zoom = min
        self._max_zoom = max

    ##  Set the point around which the camera rotates
    #
    #   \param origin type(Vector) origin point
    def setOrigin(self, origin):
        translation = origin - self._origin
        self._origin = origin
        self._scene.getActiveCamera().translate(translation)
        self._rotateCamera(0.0, 0.0)

    ##  Get the point around which the camera rotates
    #
    #   \return type(Vector) origin point
    def getOrigin(self):
        return self._origin

    ##  Prepare modifier-key variables on each event
    #
    #   \param event type(Event) event passed from event handler
    def checkModifierKeys(self, event):
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        self._shift_is_active = modifiers == QtCore.Qt.ShiftModifier
        self._ctrl_is_active = modifiers == QtCore.Qt.ControlModifier
        # Checks for the press and release event of the space key
        if event.type is Event.KeyPressEvent:
            if event.key == KeyEvent.SpaceKey:
                self._space_is_active = True
        if event.type is Event.KeyReleaseEvent:
            if event.key == KeyEvent.SpaceKey:
                self._space_is_active = False

    ##  Check if the event warrants a call off the _moveCamera method
    #
    #   \param event type(Event) event passed from event handler
    #   \return type(boolean)
    def moveEvent(self, event):
        if MouseEvent.MiddleButton in event.buttons:  # mousewheel
            return True
        elif MouseEvent.LeftButton in event.buttons and self._shift_is_active is True:  # shift -> leftbutton
            return True
        elif MouseEvent.RightButton in event.buttons and self._shift_is_active is True:  # shift -> rightbutton
            return True

    ##  Check if the event warrants a call off the _rotateCamera method
    #
    #   \param event type(Event) event passed from event handler
    #   \return type(boolean)
    def rotateEvent(self, event):
        if MouseEvent.RightButton in event.buttons:  # rightbutton
            return True
        elif MouseEvent.LeftButton in event.buttons and self._space_is_active is True:  # shift -> leftbutton
            return True

    ##  Calls the zoomaction method for the mousewheel event, mouseMoveEvent (in combo with alt or space) and when the plus or minus keys are used
    #
    #   \param event type(Event) event passed from event handler
    #   \return type(boolean)
    def initiateZoom(self, event):
        if event.type is Event.MousePressEvent:
            return False
        elif event.type is Event.MouseMoveEvent and self._space_is_active is True:  # space -> mousemove
            if self._start_y is None:
                self._start_y = 0
                _diff_y = self._start_y - event.y
            if _diff_y != 0.0:
                _zoom_speed = 2000
                self._zoomCamera(_diff_y * _zoom_speed)
                self._start_y = None
            return True

        elif event.type is Event.MouseWheelEvent:
            self._zoomCamera(event.vertical)
            return True

        elif event.type is Event.KeyPressEvent:
            if event.key == KeyEvent.MinusKey or event.key == KeyEvent.UnderscoreKey:  # checks for both the minus and underscore key because they usually share a button on the keyboard and are sometimes interchanged
                self._zoomCamera(-self._manual_zoom)
                return True
            elif event.key == KeyEvent.PlusKey or event.key == KeyEvent.EqualKey:  # same story as the minus and underscore key: it checks for both the plus and equal key (so you won't have to do shift -> equal, to use the plus-key)
                self._zoomCamera(self._manual_zoom)
                return True

    ##  Handle mouse and keyboard events
    #
    #   \param event type(Event)
    def event(self, event):
        self.checkModifierKeys(event)
        # Handle mouse- and keyboard-initiated zoom-events
        if self.initiateZoom(event):
            return True

        if(self.application.getController().getActiveView() and not self.application.getController().getActiveView()._takeCareOfKeyboard):
        # Handle keyboard-initiated rotate-events
            if event.type is event.KeyPressEvent and not self._ctrl_is_active:
                if event.key == KeyEvent.UpKey:
                    self._rotateCamera(0, 0.01)
                if event.key == KeyEvent.DownKey:
                    self._rotateCamera(0, -0.01)
                if event.key == KeyEvent.RightKey:
                    self._rotateCamera(-0.01, 0)
                if event.key == KeyEvent.LeftKey:
                    self._rotateCamera(0.01, 0)

        # Handle mouse-initiated rotate- and move-events
        if event.type is Event.MousePressEvent:
            if self.moveEvent(event) == True:
                self._move = True
                self._start_drag = (event.x, event.y)
                return False
            elif self.rotateEvent(event) == True:
                self._rotate = True
                self._start_drag = (event.x, event.y)
                return True

        elif event.type is Event.MouseMoveEvent:
            if self._rotate or self._move:
                diff = (event.x - self._start_drag[0], event.y - self._start_drag[1])
                length_squared = diff[0] * diff[0] + diff[1] * diff[1]

                if length_squared > (self._drag_distance * self._drag_distance):
                    if self._rotate:
                        self._rotateCamera(event.deltaX, event.deltaY)
                        self._dragged = True
                        return True
                    elif self._move:
                        self._moveCamera(event)
                        self._dragged = True
                        return True

        elif event.type is Event.MouseReleaseEvent:
            if self._rotate or self._move:
                self._rotate = False
                self._move = False
                self._start_drag = None
            if self._dragged:
                self._dragged = False
                if MouseEvent.RightButton in event.buttons:
                    return True

        return False

    ##  Move the camera in response to a mouse event.
    #
    #   \param event type(Event) event passed from event handler
    def _moveCamera(self, event):
        camera = self._scene.getActiveCamera()
        if not camera or not camera.isEnabled():
            return

        self._scene.acquireLock()

        camera_position = camera.getWorldPosition()
        camera.translate(Vector(-event.deltaX * 100, event.deltaY * 100, 0))
        translation = camera.getWorldPosition() - camera_position
        self._origin += translation

        self._scene.releaseLock()

    def _calc3Dto2D(self,position:numpy.array,projection:Matrix,transformation:Matrix,ortho:bool=False)-> tuple:


        #Convert to format  [x,y,z,1.0]
        if len(position) < 4:
            positionWithOne = numpy.ones((4), dtype=numpy.float32)
            positionWithOne[0:3] = position
            position=positionWithOne

        invTrans = transformation.getInverse()

        position = invTrans.getData().dot(position)
        position = projection.getData().dot(position)


        divider = 1 if ortho else position[2] * 2.0
        return position[0] /divider, position[1] / divider , position[2]

    # We need to know how big is one cm set at 0,0 position, on the screen
    # is is helpful while we are trying to estimate how small can be table object
    def _calcDistanceBeetweenTwoPoints(self,leftPoint:numpy.array,rightPoint:numpy.array,projection:Matrix,transformation:Matrix,ortho:bool=False)->numpy.array:

        # leftPoint = Vector(leftPoint[0], leftPoint[1], leftPoint[2])
        # rightPoint = Vector(rightPoint[0], rightPoint[1], rightPoint[2])

        upperPoint = numpy.array([leftPoint[0], leftPoint[1], -rightPoint[0]], dtype=numpy.float32)

        leftPointOnScreen = self._calc3Dto2D(leftPoint,projection,transformation,ortho)
        rightPointOnScreen = self._calc3Dto2D(rightPoint,projection,transformation,ortho)
        upperPointOnScreen = self._calc3Dto2D(upperPoint,projection,transformation,ortho)

        #calculatedSize = abs(leftPointOnScreen[0] - rightPointOnScreen[0])

        leftX = leftPointOnScreen[0]+0.5
        leftY = leftPointOnScreen[1]+0.5
        rightX = rightPointOnScreen[0]+0.5
        rightY = rightPointOnScreen[1]+0.5

        upperX = upperPointOnScreen[0]+0.5
        upperY= upperPointOnScreen[1]+0.5


        calculatedSize = math.sqrt((leftX - rightX)**2 + (leftY-rightY)**2)
        calculatedSizeY = math.sqrt((leftX - upperX)**2 + (leftY-upperY)**2)

        calculatedSize = calculatedSize if calculatedSize > calculatedSizeY   else calculatedSizeY

        return calculatedSize



    def _setZoomForOrtho(self,leftPoint,rightPoint,maxSize,minSize,transformation,camera,zoom_range)->(float,float):
        """

        This is a function which set zoom for current zoom_range in orthographic mode.

        :param leftPoint: Numpy witho position of left point in world coordinates
        :param rightPoint: ---
        :param maxSize: Maximum distance between point in screen coordinates
        :param minSize: ---
        :param transformation: Camera transformation matrix
        :param camera: ---
        :param zoom_range: Amount of zoom passed from mouse
        :return:
            returns calculated zoom(which is later used in ortho projection calculation) ,
            and achieved distance between points on screen

            When there is an error, function returns 0,0
        """
        zoom = camera.getZoom()
        # 0.5 we want to move slower
        zoom += -zoom_range * 0.05
        distance=0
        try:
            projectionMatrix = self.application.getMainWindow().getOrthoProjection(zoom);
            distance = self._calcDistanceBeetweenTwoPoints(leftPoint, rightPoint, projectionMatrix, transformation,
                                                           True)
            if (zoom_range >= 0):
                # We are comming closer
                if distance >= maxSize:
                    return 0,0
            else:
                if distance <= minSize:
                    return 0,0

            camera.setZoom(zoom)

        except ZeroDivisionError:
            return 0,0

        return zoom, distance

    # Make sure that ortho distance between point match provided currentDistane
    def _resizeVieportTillViewsAreCommon(self,currentDistance,leftPoint,rightPoint,maxsize,minsize,transformation,camera,initialvalue):
        """

        This function is resizing viewport as long as the distance beetween points on
        screen(in ortho) doesn't match distance between tha same point in perspective mode.


        :param currentDistance: Current distance between points in screen coordinates in perspective mode
        :param leftPoint: Left point in world coordinates
        :param rightPoint: --
        :param maxsize: Max distance between points in screen coordinates
        :param minsize: --
        :param transformation: Camerea transformation
        :param camera: --
        :param initialvalue: zoom range value
        :return:
        """


        zoom, distanceBetweenPoints = self._setZoomForOrtho(leftPoint, rightPoint, maxsize,
                                                            minsize, transformation,
                                                            camera, initialvalue)

        self._mutex.acquire(False)
        resizeStep = 2
        sign = 1

        while (distanceBetweenPoints < currentDistance):
            resizeStep += 1 * sign
            lastDistanceBetweenPoints = distanceBetweenPoints
            zoom, distanceBetweenPoints = self._setZoomForOrtho(self._leftPoint, rightPoint, maxsize,
                                                                minsize,
                                                                transformation, camera, resizeStep)
            if(zoom == 0 and distanceBetweenPoints == 0):
                break
            if(lastDistanceBetweenPoints>distanceBetweenPoints):
                sign = -1

        resizeStep = 2
        sign = -1

        while (distanceBetweenPoints > currentDistance):
            resizeStep += 1 * sign
            lastDistanceBetweenPoints = distanceBetweenPoints
            zoom, distanceBetweenPoints = self._setZoomForOrtho(self._leftPoint, rightPoint, maxsize,
                                                                minsize,
                                                                transformation, camera, resizeStep)

        self._mutex.release()


    # Make sure that perspective distance between points match provided  currentDistance

    def _moveCameraTillViewsAreCommon(self,distanceBetweenPoints,leftPoint,rightPoint,maxSize,minSize,transformation,camera,zoom_range):

        sX, sY, sZ, currentDistance, trans = self._calculateTranslationForZoom(leftPoint, rightPoint, maxSize, minSize,
                                                                               transformation, camera, 0)

        wasSmaller = False
        resizeStep = 1
        iter = 0
        while currentDistance < distanceBetweenPoints:

            resizeStep += math.sqrt(zoom_range ** 2)
            prevCurDist = currentDistance
            sX, sY, sZ, currentDistance, trans = self._calculateTranslationForZoom(leftPoint, rightPoint, maxSize,
                                                                                   minSize, transformation, camera,
                                                                                   resizeStep)
            wasSmaller = True

            iter += 1
            if (iter > 2):
                break
            if (currentDistance < prevCurDist):
                break

        resizeStep = 1
        iter = 0
        while currentDistance > distanceBetweenPoints and not wasSmaller:

            resizeStep -= math.sqrt(zoom_range ** 2)
            prevCurDist = currentDistance
            sX, sY, sZ, currentDistance, trans = self._calculateTranslationForZoom(leftPoint, rightPoint, maxSize,
                                                                                   minSize, transformation, camera,
                                                                                   resizeStep)

            if (currentDistance > prevCurDist):
                break
            iter += 1
            if (iter > 2):
                break

        if (sZ != 0):
            camera.translate(Vector(sX, sY, sZ))

        self.application.getMainWindow().updateViewportGeometry()

    def setToTheSameDistanceinOrth(self,tableWidth):
        camera = self._scene.getActiveCamera()
        if not camera or not camera.isEnabled():
            return 0

        self._scene.acquireLock()
        transformation = camera.getWorldTransformation()
        rightPoint = Vector(tableWidth, 0, 0)
        perspective = camera.getPerspectiveMatrix()
        # if not perspective:
        #     perspective = camera.getProjectionMatrix()
            
        currentDistance = self._calcDistanceBeetweenTwoPoints(self._leftPoint, self._rightPoint, perspective,
                         transformation)

        self._resizeVieportTillViewsAreCommon(currentDistance,self._leftPoint, self._rightPoint, self._max_size, self._min_size, transformation, camera, 0)

        if not camera.isPerspective():
            self.application.getMainWindow().updateViewportGeometry()
            camera.transformationChanged.emit(camera)

        self._scene.releaseLock()

        return currentDistance

    def _calculateTranslationForZoom(self,left,right,max,min,transformation,camera,zoom_range) -> (float,float,float):
        """
        :param left: leftPoint numpy
        :param right: rightPoint
        :param max: maximum distance beetween points on screen
        :param min: minimum distance
        :param proj: projection matrix
        :param trans: transformation matrix
        :param camera:
        :param zoom_range: value received from mouse scroll
        :return: (shiftX,shiftY,shiftZ) tuple with values of translation
        """
        shift=shiftx=shifty=0;

        #
        # zDistance = 10 * camera.getPerspectiveMatrix().getData()[1][1] / distanceBetweenPoints / 3
        # v = camera.getPosition()
        # currentPosition = Quaternion(v.x,v.y,v.z,0)
        # currentPosition = camera.getOrientation().toMatrix().getData().dot(currentPosition.getData())
        # currentZ = abs(currentPosition[2])
        #
        # diff = abs(zDistance - currentZ)
        # if(currentZ > zDistance):
        #     diff*=-1
        # transform = transformation.getData()
        # transform[2][3]+diff;
        # transformation=Matrix(transform)
        # camera.setTransformation(transformation)
        data = transformation.getData()

        point = Vector(data[0][3],data[1][3],data[2][3])

        zK = transformation.getData()[2][3]
        transData = transformation.getData()
        r = (camera.getWorldPosition() - Vector(left)).length()

        delta = r * (zoom_range / 128 / 10.0)
        delta *= -1

        shift = 0.1 * delta
        shiftX = 0 * delta
        shiftY = 0 * delta

        vector = Vector(shiftx,shifty,shift)

        # Check if which half of coordinate system camera is placed
        if (zK <= 0):
            transData[0][3] -= shiftx
            transData[1][3] -= shifty
            transData[2][3] -= shift
        if (zK > 0):
            transData[0][3] += shiftx
            transData[1][3] += shifty
            transData[2][3] += shift
        transformation = Matrix(transData)
        calculatedSize = self._calcDistanceBeetweenTwoPoints(left, right, camera.getPerspectiveMatrix(),
                                                             transformation)


        if zoom_range > 0 and calculatedSize > max:
            shift = shiftX = shiftY = 0
        elif zoom_range < 0 and calculatedSize < min:
            shift = shiftX = shiftY = 0

        return shiftx, shifty, shift, calculatedSize, transformation

    ##  "Zoom" the camera in response to a mouse event.
    #
    #   \param zoom_range type(int) factor by which the distance to the origin is multiplied, multiplied by 1280
    def _zoomCamera(self, zoom_range):

        zoom_range*=self._zoom_speed
        camera = self._scene.getActiveCamera()
        if not camera or not camera.isEnabled() or zoom_range==0.:
            return

        maxSize = self._max_size
        minSize = self._min_size
        leftPoint = self._leftPoint
        rightPoint = self._rightPoint
        #upperPoint = numpy.array([0, 0, -10.], dtype=numpy.float32)
        #rightPoint += leftPoint


        # leftPoint = self.getOrigin().getData()
        # camera.lookAt(Vector(leftPoint))
        # camera.lookAt(self._origin)
        # --------------------------------LOCK-----------------------------
        self._scene.acquireLock()
        # self._mutex.acquire(False)
        transformation = camera.getWorldTransformation()

        sX=sY=sZ=0
        if not camera.isPerspective():
            # --------------------------------------------------------- ORTHOGRAPHIC

            zoom, distanceBetweenPoints = self._setZoomForOrtho(leftPoint,rightPoint,maxSize,minSize,transformation,camera,zoom_range)

            if(zoom!=0):

                self._moveCameraTillViewsAreCommon(distanceBetweenPoints,leftPoint,rightPoint,maxSize,minSize,transformation,camera,zoom_range)

        else:
            # ---------------------------------------------------------- PERSPECTIVE

            zoom_range*=2
            sX,sY,sZ,currentDistance,trans = self._calculateTranslationForZoom(leftPoint,rightPoint,maxSize,minSize,transformation,camera,zoom_range)

            if sZ!=0:
                camera.translate(Vector(sX, sY, sZ))

            self._resizeVieportTillViewsAreCommon(currentDistance,leftPoint,rightPoint,maxSize,minSize,transformation,camera,zoom_range)


        self._scene.releaseLock()
        camera.transformationChanged.emit(camera)

        # self._mutex.release()

    ##  Rotate the camera in response to a mouse event.
    #
    #   \param x type(int) amount by which the camera should be rotated horizontally, expressed in pixelunits
    #   \param y type(int) amount by which the camera should be rotated vertically, expressed in pixelunits
    def _rotateCamera(self, x, y):
        camera = self._scene.getActiveCamera()
        if not camera or not camera.isEnabled():
            return

        self._scene.acquireLock()

        dx = math.radians(x * 180.0)
        dy = math.radians(y * 180.0)

        diff = camera.getPosition() - self._origin
        my = Matrix()
        my.setByRotationAxis(dx, Vector.Unit_Y)

        mx = Matrix(my.getData())
        mx.rotateByAxis(dy, Vector.Unit_Y.cross(diff).normalized())

        n = diff.multiply(mx)

        try:
            angle = math.acos(Vector.Unit_Y.dot(n.normalized()))
        except ValueError:
            return

        if angle < 0.1 or angle > math.pi - 0.1:
            n = diff.multiply(my)

        n += self._origin

        camera.setPosition(n)
        camera.lookAt(self._origin)
        
        self._scene.releaseLock()
